@@ -1,6 +1,6 @@
 /* ----------------------------------------------------------------------- *
- *   
- *   Copyright 1996-2009 The NASM Authors - All Rights Reserved
+ *
+ *   Copyright 1996-2010 The NASM Authors - All Rights Reserved
  *   See the file AUTHORS included with the NASM distribution for
  *   the specific copyright holders.
  *
@@ -14,7 +14,7 @@
  *     copyright notice, this list of conditions and the following
  *     disclaimer in the documentation and/or other materials provided
  *     with the distribution.
- *     
+ *
  *     THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
  *     CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  *     INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
@@ -31,9 +31,9 @@
  *
  * ----------------------------------------------------------------------- */
 
-/* 
- * outcoff.c	output routines for the Netwide Assembler to produce
- *		COFF object files (for DJGPP and Win32)
+/*
+ * outcoff.c    output routines for the Netwide Assembler to produce
+ *              COFF object files (for DJGPP and Win32)
  */
 
 #include "compiler.h"
@@ -49,8 +49,10 @@
 #include "nasmlib.h"
 #include "saa.h"
 #include "raa.h"
+#include "eval.h"
 #include "output/outform.h"
 #include "output/outlib.h"
+#include "output/pecoff.h"
 
 #if defined(OF_COFF) || defined(OF_WIN32) || defined(OF_WIN64)
 
@@ -105,8 +107,8 @@ static int32_t imagebase_sect;
 
 struct Reloc {
     struct Reloc *next;
-    int32_t address;		/* relative to _start_ of section */
-    int32_t symbol;		/* symbol number */
+    int32_t address;            /* relative to _start_ of section */
+    int32_t symbol;             /* symbol number */
     enum {
         SECT_SYMBOLS,
         ABS_SYMBOL,
@@ -115,28 +117,17 @@ struct Reloc {
     int16_t type;
 };
 
-/* possible values for Reloc->type */
-#define IMAGE_REL_AMD64_ADDR64		0x0001
-#define IMAGE_REL_AMD64_ADDR32		0x0002
-#define IMAGE_REL_AMD64_ADDR32NB	0x0003
-#define IMAGE_REL_AMD64_REL32		0x0004
-#define IMAGE_REL_I386_DIR32		0x0006
-#define IMAGE_REL_I386_DIR32NB		0x0007
-#define IMAGE_REL_I386_REL32		0x0014
-
 struct Symbol {
     char name[9];
-    int32_t strpos;		/* string table position of name */
-    int32_t value;		/* address, or COMMON variable size */
+    int32_t strpos;             /* string table position of name */
+    int32_t value;              /* address, or COMMON variable size */
     int section;                /* section number where it's defined
                                  * - in COFF codes, not NASM codes */
-    bool is_global;              /* is it a global symbol or not? */
-    int16_t type;		/* 0 - notype, 0x20 - function */
-    int32_t namlen;		/* full name length */
+    bool is_global;             /* is it a global symbol or not? */
+    int16_t type;               /* 0 - notype, 0x20 - function */
+    int32_t namlen;             /* full name length */
 };
 
-static FILE *coffp;
-static efunc error;
 static char coff_infile[FILENAME_MAX];
 
 struct Section {
@@ -145,16 +136,65 @@ struct Section {
     int nrelocs;
     int32_t index;
     struct Reloc *head, **tail;
-    uint32_t flags;        /* section flags */
+    uint32_t flags;             /* section flags */
     char name[9];
     int32_t pos, relpos;
 };
 
-#define TEXT_FLAGS ((win32 | win64) ? 0x60500020L : 0x20L)
-#define DATA_FLAGS ((win32 | win64) ? 0xC0300040L : 0x40L)
-#define BSS_FLAGS ((win32 | win64) ? 0xC0300080L : 0x80L)
-#define INFO_FLAGS 0x00100A00L
-#define RDATA_FLAGS ((win32 | win64) ? 0x40400040L : 0x40L)
+/*
+ * Some common section flags by default
+ */
+#define TEXT_FLAGS_WIN                                  \
+        (IMAGE_SCN_CNT_CODE                     |       \
+         IMAGE_SCN_ALIGN_16BYTES                |       \
+         IMAGE_SCN_MEM_EXECUTE                  |       \
+         IMAGE_SCN_MEM_READ)
+#define TEXT_FLAGS_DOS                                  \
+        (IMAGE_SCN_CNT_CODE)
+
+#define DATA_FLAGS_WIN                                  \
+        (IMAGE_SCN_CNT_INITIALIZED_DATA         |       \
+         IMAGE_SCN_ALIGN_4BYTES                 |       \
+         IMAGE_SCN_MEM_READ                     |       \
+         IMAGE_SCN_MEM_WRITE)
+#define DATA_FLAGS_DOS                                  \
+        (IMAGE_SCN_CNT_INITIALIZED_DATA)
+
+#define BSS_FLAGS_WIN                                   \
+        (IMAGE_SCN_CNT_UNINITIALIZED_DATA       |       \
+         IMAGE_SCN_ALIGN_4BYTES                 |       \
+         IMAGE_SCN_MEM_READ                     |       \
+         IMAGE_SCN_MEM_WRITE)
+#define BSS_FLAGS_DOS                                   \
+        (IMAGE_SCN_CNT_UNINITIALIZED_DATA)
+
+#define RDATA_FLAGS_WIN                                 \
+        (IMAGE_SCN_CNT_INITIALIZED_DATA         |       \
+         IMAGE_SCN_ALIGN_8BYTES                 |       \
+         IMAGE_SCN_MEM_READ)
+
+#define RDATA_FLAGS_DOS                                 \
+        (IMAGE_SCN_CNT_INITIALIZED_DATA)
+
+#define PDATA_FLAGS                                     \
+        (IMAGE_SCN_CNT_INITIALIZED_DATA         |       \
+         IMAGE_SCN_ALIGN_4BYTES                 |       \
+         IMAGE_SCN_MEM_READ)
+
+#define XDATA_FLAGS                                     \
+        (IMAGE_SCN_CNT_INITIALIZED_DATA         |       \
+         IMAGE_SCN_ALIGN_8BYTES                 |       \
+         IMAGE_SCN_MEM_READ)
+
+#define INFO_FLAGS                                      \
+        (IMAGE_SCN_ALIGN_1BYTES                 |       \
+         IMAGE_SCN_LNK_INFO                     |       \
+         IMAGE_SCN_LNK_REMOVE)
+
+#define TEXT_FLAGS      ((win32 | win64) ? TEXT_FLAGS_WIN  : TEXT_FLAGS_DOS)
+#define DATA_FLAGS      ((win32 | win64) ? DATA_FLAGS_WIN  : DATA_FLAGS_DOS)
+#define BSS_FLAGS       ((win32 | win64) ? BSS_FLAGS_WIN   : BSS_FLAGS_DOS)
+#define RDATA_FLAGS     ((win32 | win64) ? RDATA_FLAGS_WIN : RDATA_FLAGS_DOS)
 
 #define SECT_DELTA 32
 static struct Section **sects;
@@ -172,51 +212,39 @@ static struct RAA *bsym, *symval;
 static struct SAA *strs;
 static uint32_t strslen;
 
-static void coff_gen_init(FILE *, efunc);
-static void coff_sect_write(struct Section *, const uint8_t *,
-                            uint32_t);
+static void coff_gen_init(void);
+static void coff_sect_write(struct Section *, const uint8_t *, uint32_t);
 static void coff_write(void);
 static void coff_section_header(char *, int32_t, int32_t, int32_t, int32_t, int, int32_t);
 static void coff_write_relocs(struct Section *);
 static void coff_write_symbols(void);
 
-static void coff_win32_init(FILE * fp, efunc errfunc,
-                            ldfunc ldef, evalfunc eval)
+static void coff_win32_init(void)
 {
-    win32 = true; win64 = false;
-    (void)ldef;                 /* placate optimizers */
-    (void)eval;
-    coff_gen_init(fp, errfunc);
+    win32 = true;
+    win64 = false;
+    coff_gen_init();
 }
 
-static void coff_win64_init(FILE * fp, efunc errfunc,
-                            ldfunc ldef, evalfunc eval)
+static void coff_win64_init(void)
 {
-    extern struct ofmt of_win64;
-
     maxbits = 64;
-    win32 = false; win64 = true;
-    (void)ldef;                 /* placate optimizers */
-    (void)eval;
-    coff_gen_init(fp, errfunc);
+    win32 = false;
+    win64 = true;
+    coff_gen_init();
     imagebase_sect = seg_alloc()+1;
-    ldef(WRT_IMAGEBASE,imagebase_sect,0,NULL,false,false,&of_win64,errfunc);
+    define_label(WRT_IMAGEBASE, imagebase_sect, 0, NULL, false, false);
 }
 
-static void coff_std_init(FILE * fp, efunc errfunc, ldfunc ldef,
-                          evalfunc eval)
+static void coff_std_init(void)
 {
     win32 = win64 = false;
-    (void)ldef;                 /* placate optimizers */
-    (void)eval;
-    coff_gen_init(fp, errfunc);
+    coff_gen_init();
 }
 
-static void coff_gen_init(FILE * fp, efunc errfunc)
+static void coff_gen_init(void)
 {
 
-    coffp = fp;
-    error = errfunc;
     sects = NULL;
     nsects = sectlen = 0;
     syms = saa_init(sizeof(struct Symbol));
@@ -236,7 +264,6 @@ static void coff_cleanup(int debuginfo)
     (void)debuginfo;
 
     coff_write();
-    fclose(coffp);
     for (i = 0; i < nsects; i++) {
         if (sects[i]->data)
             saa_free(sects[i]->data);
@@ -277,7 +304,7 @@ static int coff_make_section(char *name, uint32_t flags)
     s->flags = flags;
 
     if (nsects >= sectlen) {
-	sectlen += SECT_DELTA;
+        sectlen += SECT_DELTA;
         sects = nasm_realloc(sects, sectlen * sizeof(*sects));
     }
     sects[nsects++] = s;
@@ -285,6 +312,11 @@ static int coff_make_section(char *name, uint32_t flags)
     return nsects - 1;
 }
 
+static inline int32_t coff_sectalign_flags(unsigned int align)
+{
+    return (ilog2_32(align) + 1) << 20;
+}
+
 static int32_t coff_section_names(char *name, int pass, int *bits)
 {
     char *p;
@@ -310,7 +342,7 @@ static int32_t coff_section_names(char *name, int pass, int *bits)
     if (*p)
         *p++ = '\0';
     if (strlen(name) > 8) {
-        error(ERR_WARNING, "COFF section names limited to 8 characters:"
+        nasm_error(ERR_WARNING, "COFF section names limited to 8 characters:"
               " truncating");
         name[8] = '\0';
     }
@@ -336,7 +368,7 @@ static int32_t coff_section_names(char *name, int pass, int *bits)
                 flags = RDATA_FLAGS;
             else {
                 flags = DATA_FLAGS;     /* gotta do something */
-                error(ERR_NONFATAL, "standard COFF does not support"
+                nasm_error(ERR_NONFATAL, "standard COFF does not support"
                       " read-only data sections");
             }
         } else if (!nasm_stricmp(q, "bss")) {
@@ -346,34 +378,28 @@ static int32_t coff_section_names(char *name, int pass, int *bits)
                 flags = INFO_FLAGS;
             else {
                 flags = DATA_FLAGS;     /* gotta do something */
-                error(ERR_NONFATAL, "standard COFF does not support"
+                nasm_error(ERR_NONFATAL, "standard COFF does not support"
                       " informational sections");
             }
         } else if (!nasm_strnicmp(q, "align=", 6)) {
             if (!(win32 | win64))
-                error(ERR_NONFATAL, "standard COFF does not support"
+                nasm_error(ERR_NONFATAL, "standard COFF does not support"
                       " section alignment specification");
             else {
                 if (q[6 + strspn(q + 6, "0123456789")])
-                    error(ERR_NONFATAL,
+                    nasm_error(ERR_NONFATAL,
                           "argument to `align' is not numeric");
                 else {
                     unsigned int align = atoi(q + 6);
                     if (!align || ((align - 1) & align))
-                        error(ERR_NONFATAL, "argument to `align' is not a"
+                        nasm_error(ERR_NONFATAL, "argument to `align' is not a"
                               " power of two");
                     else if (align > 64)
-                        error(ERR_NONFATAL, "Win32 cannot align sections"
+                        nasm_error(ERR_NONFATAL, "Win32 cannot align sections"
                               " to better than 64-byte boundaries");
                     else {
                         align_and = ~0x00F00000L;
-                        align_or = (align == 1 ? 0x00100000L :
-                                    align == 2 ? 0x00200000L :
-                                    align == 4 ? 0x00300000L :
-                                    align == 8 ? 0x00400000L :
-                                    align == 16 ? 0x00500000L :
-                                    align ==
-                                    32 ? 0x00600000L : 0x00700000L);
+                        align_or  = coff_sectalign_flags(align);
                     }
                 }
             }
@@ -391,10 +417,10 @@ static int32_t coff_section_names(char *name, int pass, int *bits)
                 flags = RDATA_FLAGS;
             else if (!strcmp(name, ".bss"))
                 flags = BSS_FLAGS;
-	    else if (win64 && !strcmp(name, ".pdata"))
-		flags = 0x40300040; /* rdata align=4 */
-	    else if (win64 && !strcmp(name, ".xdata"))
-		flags = 0x40400040; /* rdate align=8 */
+            else if (win64 && !strcmp(name, ".pdata"))
+                flags = PDATA_FLAGS;
+            else if (win64 && !strcmp(name, ".xdata"))
+                flags = XDATA_FLAGS;
             else
                 flags = TEXT_FLAGS;
         }
@@ -405,7 +431,7 @@ static int32_t coff_section_names(char *name, int pass, int *bits)
         sects[i]->flags |= align_or;
     } else if (pass == 1) {
         if (flags)
-            error(ERR_WARNING, "section attributes ignored on"
+            nasm_error(ERR_WARNING, "section attributes ignored on"
                   " redeclaration of section `%s'", name);
     }
 
@@ -419,19 +445,19 @@ static void coff_deflabel(char *name, int32_t segment, int64_t offset,
     struct Symbol *sym;
 
     if (special)
-        error(ERR_NONFATAL, "COFF format does not support any"
+        nasm_error(ERR_NONFATAL, "COFF format does not support any"
               " special symbol types");
 
     if (name[0] == '.' && name[1] == '.' && name[2] != '@') {
-	if (strcmp(name,WRT_IMAGEBASE))
-            error(ERR_NONFATAL, "unrecognized special symbol `%s'", name);
+        if (strcmp(name,WRT_IMAGEBASE))
+            nasm_error(ERR_NONFATAL, "unrecognized special symbol `%s'", name);
         return;
     }
 
     if (strlen(name) > 8) {
-	size_t nlen = strlen(name)+1;
+        size_t nlen = strlen(name)+1;
         saa_wbytes(strs, name, nlen);
-	strslen += nlen;
+        strslen += nlen;
     } else
         pos = -1;
 
@@ -442,7 +468,7 @@ static void coff_deflabel(char *name, int32_t segment, int64_t offset,
     if (pos == -1)
         strcpy(sym->name, name);
     sym->is_global = !!is_global;
-    sym->type = 0;		/* Default to T_NULL (no type) */
+    sym->type = 0;              /* Default to T_NULL (no type) */
     if (segment == NO_SEG)
         sym->section = -1;      /* absolute symbol */
     else {
@@ -465,9 +491,8 @@ static void coff_deflabel(char *name, int32_t segment, int64_t offset,
      * define the references from external-symbol segment numbers
      * to these symbol records.
      */
-    if (sym->section == 0) {
+    if (sym->section == 0)
         bsym = raa_write(bsym, segment, nsyms);
-    }
 
     if (segment != NO_SEG)
         symval = raa_write(symval, segment, sym->section ? 0 : sym->value);
@@ -476,7 +501,7 @@ static void coff_deflabel(char *name, int32_t segment, int64_t offset,
 }
 
 static int32_t coff_add_reloc(struct Section *sect, int32_t segment,
-				int16_t type)
+                              int16_t type)
 {
     struct Reloc *r;
 
@@ -485,17 +510,18 @@ static int32_t coff_add_reloc(struct Section *sect, int32_t segment,
     r->next = NULL;
 
     r->address = sect->len;
-    if (segment == NO_SEG)
+    if (segment == NO_SEG) {
         r->symbol = 0, r->symbase = ABS_SYMBOL;
-    else {
+    } else {
         int i;
         r->symbase = REAL_SYMBOLS;
-        for (i = 0; i < nsects; i++)
+        for (i = 0; i < nsects; i++) {
             if (segment == sects[i]->index) {
                 r->symbol = i * 2;
                 r->symbase = SECT_SYMBOLS;
                 break;
             }
+        }
         if (r->symbase == REAL_SYMBOLS)
             r->symbol = raa_read(bsym, segment);
     }
@@ -508,12 +534,12 @@ static int32_t coff_add_reloc(struct Section *sect, int32_t segment,
      */
     if (r->symbase == REAL_SYMBOLS && !(win32 | win64))
         return raa_read(symval, segment);
-    else
-        return 0;
+
+    return 0;
 }
 
 static void coff_out(int32_t segto, const void *data,
-		     enum out_type type, uint64_t size,
+                     enum out_type type, uint64_t size,
                      int32_t segment, int32_t wrt)
 {
     struct Section *s;
@@ -522,7 +548,7 @@ static void coff_out(int32_t segto, const void *data,
 
     if (wrt != NO_SEG && !win64) {
         wrt = NO_SEG;           /* continue to do _something_ */
-        error(ERR_NONFATAL, "WRT not supported by COFF output formats");
+        nasm_error(ERR_NONFATAL, "WRT not supported by COFF output formats");
     }
 
     /*
@@ -530,32 +556,34 @@ static void coff_out(int32_t segto, const void *data,
      */
     if (segto == NO_SEG) {
         if (type != OUT_RESERVE)
-            error(ERR_NONFATAL, "attempt to assemble code in [ABSOLUTE]"
+            nasm_error(ERR_NONFATAL, "attempt to assemble code in [ABSOLUTE]"
                   " space");
         return;
     }
 
     s = NULL;
-    for (i = 0; i < nsects; i++)
+    for (i = 0; i < nsects; i++) {
         if (segto == sects[i]->index) {
             s = sects[i];
             break;
         }
+    }
     if (!s) {
         int tempint;            /* ignored */
         if (segto != coff_section_names(".text", 2, &tempint))
-            error(ERR_PANIC, "strange segment conditions in COFF driver");
+            nasm_error(ERR_PANIC, "strange segment conditions in COFF driver");
         else
             s = sects[nsects - 1];
     }
 
     /* magically default to 'wrt ..imagebase' in .pdata and .xdata */
-    if (win64 && wrt == NO_SEG &&
-	(!strcmp(s->name,".pdata") || !strcmp(s->name,".xdata")))
-	wrt = imagebase_sect;
+    if (win64 && wrt == NO_SEG) {
+        if (!strcmp(s->name,".pdata") || !strcmp(s->name,".xdata"))
+            wrt = imagebase_sect;
+    }
 
     if (!s->data && type != OUT_RESERVE) {
-        error(ERR_WARNING, "attempt to initialize memory in"
+        nasm_error(ERR_WARNING, "attempt to initialize memory in"
               " BSS section `%s': ignored", s->name);
         s->len += realsize(type, size);
         return;
@@ -563,28 +591,28 @@ static void coff_out(int32_t segto, const void *data,
 
     if (type == OUT_RESERVE) {
         if (s->data) {
-            error(ERR_WARNING, "uninitialised space declared in"
+            nasm_error(ERR_WARNING, "uninitialised space declared in"
                   " non-BSS section `%s': zeroing", s->name);
             coff_sect_write(s, NULL, size);
         } else
             s->len += size;
     } else if (type == OUT_RAWDATA) {
         if (segment != NO_SEG)
-            error(ERR_PANIC, "OUT_RAWDATA with other than NO_SEG");
+            nasm_error(ERR_PANIC, "OUT_RAWDATA with other than NO_SEG");
         coff_sect_write(s, data, size);
     } else if (type == OUT_ADDRESS) {
         if (!(win64)) {
-            if (size != 4 && (segment != NO_SEG || wrt != NO_SEG))
-                error(ERR_NONFATAL, "COFF format does not support non-32-bit"
+            if (size != 4 && (segment != NO_SEG || wrt != NO_SEG)) {
+                nasm_error(ERR_NONFATAL, "COFF format does not support non-32-bit"
                       " relocations");
-            else {
+            } else {
                 int32_t fix = 0;
                 if (segment != NO_SEG || wrt != NO_SEG) {
                     if (wrt != NO_SEG) {
-                        error(ERR_NONFATAL, "COFF format does not support"
+                        nasm_error(ERR_NONFATAL, "COFF format does not support"
                               " WRT types");
                     } else if (segment % 2) {
-                        error(ERR_NONFATAL, "COFF format does not support"
+                        nasm_error(ERR_NONFATAL, "COFF format does not support"
                               " segment base references");
                     } else
                         fix = coff_add_reloc(s, segment, IMAGE_REL_I386_DIR32);
@@ -597,38 +625,38 @@ static void coff_out(int32_t segto, const void *data,
             int32_t fix = 0;
             p = mydata;
             if (size == 8) {
-	        if (wrt == imagebase_sect) {
-		    error(ERR_NONFATAL, "operand size mismatch: 'wrt "
-			WRT_IMAGEBASE "' is a 32-bit operand");
-		}
+                if (wrt == imagebase_sect) {
+                    nasm_error(ERR_NONFATAL, "operand size mismatch: 'wrt "
+                               WRT_IMAGEBASE "' is a 32-bit operand");
+                }
                 fix = coff_add_reloc(s, segment, IMAGE_REL_AMD64_ADDR64);
                 WRITEDLONG(p, *(int64_t *)data + fix);
                 coff_sect_write(s, mydata, size);
             } else {
                 fix = coff_add_reloc(s, segment,
-			wrt == imagebase_sect ?	IMAGE_REL_AMD64_ADDR32NB:
-						IMAGE_REL_AMD64_ADDR32);
+                        wrt == imagebase_sect ?	IMAGE_REL_AMD64_ADDR32NB:
+                                                IMAGE_REL_AMD64_ADDR32);
                 WRITELONG(p, *(int64_t *)data + fix);
                 coff_sect_write(s, mydata, size);
             }
         }
     } else if (type == OUT_REL2ADR) {
-        error(ERR_NONFATAL, "COFF format does not support 16-bit"
+        nasm_error(ERR_NONFATAL, "COFF format does not support 16-bit"
               " relocations");
     } else if (type == OUT_REL4ADR) {
         if (segment == segto && !(win64))  /* Acceptable for RIP-relative */
-            error(ERR_PANIC, "intra-segment OUT_REL4ADR");
+            nasm_error(ERR_PANIC, "intra-segment OUT_REL4ADR");
         else if (segment == NO_SEG && win32)
-            error(ERR_NONFATAL, "Win32 COFF does not correctly support"
+            nasm_error(ERR_NONFATAL, "Win32 COFF does not correctly support"
                   " relative references to absolute addresses");
         else {
             int32_t fix = 0;
             if (segment != NO_SEG && segment % 2) {
-                error(ERR_NONFATAL, "COFF format does not support"
+                nasm_error(ERR_NONFATAL, "COFF format does not support"
                       " segment base references");
             } else
                 fix = coff_add_reloc(s, segment,
-			win64 ? IMAGE_REL_AMD64_REL32 : IMAGE_REL_I386_REL32);
+                        win64 ? IMAGE_REL_AMD64_REL32 : IMAGE_REL_I386_REL32);
             p = mydata;
             if (win32 | win64) {
                 WRITELONG(p, *(int64_t *)data + 4 - size + fix);
@@ -649,7 +677,7 @@ static void coff_sect_write(struct Section *sect,
 }
 
 typedef struct tagString {
-    struct tagString *Next;
+    struct tagString *next;
     int len;
     char *String;
 } STRING;
@@ -657,9 +685,9 @@ typedef struct tagString {
 #define EXPORT_SECTION_NAME ".drectve"
 #define EXPORT_SECTION_FLAGS INFO_FLAGS
 /*
-#define EXPORT_SECTION_NAME ".text"
-#define EXPORT_SECTION_FLAGS TEXT_FLAGS
-*/
+ * #define EXPORT_SECTION_NAME ".text"
+ * #define EXPORT_SECTION_FLAGS TEXT_FLAGS
+ */
 
 static STRING *Exports = NULL;
 static struct Section *directive_sec;
@@ -669,57 +697,55 @@ void AddExport(char *name)
 
     newS = (STRING *) nasm_malloc(sizeof(STRING));
     newS->len = strlen(name);
-    newS->Next = NULL;
+    newS->next = NULL;
     newS->String = (char *)nasm_malloc(newS->len + 1);
     strcpy(newS->String, name);
     if (rvp == NULL) {
         int i;
-        for (i = 0; i < nsects; i++)
 
+        for (i = 0; i < nsects; i++) {
             if (!strcmp(EXPORT_SECTION_NAME, sects[i]->name))
                 break;
+        }
+
         if (i == nsects)
-            directive_sec =
-                sects[coff_make_section
-                      (EXPORT_SECTION_NAME, EXPORT_SECTION_FLAGS)];
-        else
-            directive_sec = sects[i];
+            i = coff_make_section(EXPORT_SECTION_NAME, EXPORT_SECTION_FLAGS);
+
+        directive_sec = sects[i];
         Exports = newS;
     } else {
-        while (rvp->Next) {
+        while (rvp->next) {
             if (!strcmp(rvp->String, name))
                 return;
-            rvp = rvp->Next;
+            rvp = rvp->next;
         }
-        rvp->Next = newS;
+        rvp->next = newS;
     }
 }
 
-void BuildExportTable(void)
+static void BuildExportTable(STRING **rvp)
 {
-    STRING *rvp = Exports, *next;
-    uint8_t buf[256];
-    int len;
-    if (rvp == NULL)
+    STRING *p, *t;
+
+    if (!rvp || !*rvp)
         return;
-    while (rvp) {
-        len = sprintf((char *)buf, "-export:%s ", rvp->String);
-        coff_sect_write(directive_sec, buf, len);
-        rvp = rvp->Next;
+
+    list_for_each_safe(p, t, *rvp) {
+        coff_sect_write(directive_sec, (uint8_t *)"-export:", 8);
+        coff_sect_write(directive_sec, (uint8_t *)p->String, p->len);
+        coff_sect_write(directive_sec, (uint8_t *)" ", 1);
+        nasm_free(p->String);
+        nasm_free(p);
     }
 
-    next = Exports;
-    while ((rvp = next)) {
-        next = rvp->Next;
-        nasm_free(rvp->String);
-        nasm_free(rvp);
-    }
-    Exports = NULL;
+    *rvp = NULL;
 }
 
-static int coff_directives(char *directive, char *value, int pass)
+static int coff_directives(enum directives directive, char *value, int pass)
 {
-    if (!strcmp(directive, "export")) {
+    switch (directive) {
+    case D_EXPORT:
+    {
         char *q, *name;
 
         if (pass == 2)
@@ -734,67 +760,81 @@ static int coff_directives(char *directive, char *value, int pass)
         }
 
         if (!*name) {
-            error(ERR_NONFATAL, "`export' directive requires export name");
+            nasm_error(ERR_NONFATAL, "`export' directive requires export name");
             return 1;
         }
         if (*q) {
-            error(ERR_NONFATAL, "unrecognized export qualifier `%s'", q);
+            nasm_error(ERR_NONFATAL, "unrecognized export qualifier `%s'", q);
             return 1;
         }
         AddExport(name);
         return 1;
-    } else if (win32 && !strcmp(directive,"safeseh")) {
-	static int sxseg=-1;
-	int i;
-	if (sxseg==-1)
-	{   for (i = 0; i < nsects; i++)
-		if (!strcmp(".sxdata",sects[i]->name))
-		    break;
-	    if (i == nsects)
-		sxseg = coff_make_section(".sxdata",0x200);
-	    else
-		sxseg = i;
-	}
-	/* pass0 == 2 is the only time when the full set of symbols are
-	   guaranteed to be present; it is the final output pass. */
-	if (pass0 == 2) {
-	    uint32_t n;
-	    saa_rewind(syms);
-	    for (n = 0; n < nsyms; n++) {
-		struct Symbol *sym = saa_rstruct(syms);
-		bool equals;
-
-		/* sym->strpos is biased by 4, because symbol
-		 * table is prefixed with table length */
-		if (sym->strpos >=4) {
-		    char *name = nasm_malloc(sym->namlen+1);
-		    saa_fread(strs, sym->strpos-4, name, sym->namlen);
-		    name[sym->namlen] = '\0';
-		    equals = !strcmp(value,name);
-		    nasm_free(name);
-		}
-		else
-		    equals = !strcmp(value,sym->name);
-
-		if (equals) {
-		    /* this value arithmetics effectively reflects
-		     * initsym in coff_write(): 2 for file, 1 for
-		     * .absolute and two per each section */
-		    unsigned char value[4],*p=value;
-		    WRITELONG(p,n + 2 + 1 + nsects*2);
-		    coff_sect_write(sects[sxseg],value,4);
-		    sym->type = 0x20;
-		    break;
-		}
-	    }
-	    if (n == nsyms) {
-		error(ERR_NONFATAL,
-		      "`safeseh' directive requires valid symbol");
-	    }
-	}
-	return 1;
     }
-    return 0;
+    case D_SAFESEH:
+    {
+        static int sxseg=-1;
+        int i;
+
+        if (!win32) /* Only applicable for -f win32 */
+            return 0;
+
+        if (sxseg == -1) {
+            for (i = 0; i < nsects; i++)
+                if (!strcmp(".sxdata",sects[i]->name))
+                    break;
+            if (i == nsects)
+                sxseg = coff_make_section(".sxdata", IMAGE_SCN_LNK_INFO);
+            else
+                sxseg = i;
+        }
+        /*
+         * pass0 == 2 is the only time when the full set of symbols are
+         * guaranteed to be present; it is the final output pass.
+         */
+        if (pass0 == 2) {
+            uint32_t n;
+            saa_rewind(syms);
+            for (n = 0; n < nsyms; n++) {
+                struct Symbol *sym = saa_rstruct(syms);
+                bool equals;
+
+                /*
+                 * sym->strpos is biased by 4, because symbol
+                 * table is prefixed with table length
+                 */
+                if (sym->strpos >=4) {
+                    char *name = nasm_malloc(sym->namlen+1);
+                    saa_fread(strs, sym->strpos-4, name, sym->namlen);
+                    name[sym->namlen] = '\0';
+                    equals = !strcmp(value,name);
+                    nasm_free(name);
+                } else {
+                    equals = !strcmp(value,sym->name);
+                }
+
+                if (equals) {
+                    /*
+                     * this value arithmetics effectively reflects
+                     * initsym in coff_write(): 2 for file, 1 for
+                     * .absolute and two per each section
+                     */
+                    unsigned char value[4],*p=value;
+                    WRITELONG(p,n + 2 + 1 + nsects*2);
+                    coff_sect_write(sects[sxseg],value,4);
+                    sym->type = 0x20;
+                    break;
+                }
+            }
+            if (n == nsyms) {
+                nasm_error(ERR_NONFATAL,
+                           "`safeseh' directive requires valid symbol");
+            }
+        }
+        return 1;
+    }
+    default:
+        return 0;
+    }
 }
 
 static void coff_write(void)
@@ -802,20 +842,21 @@ static void coff_write(void)
     int32_t pos, sympos, vsize;
     int i;
 
-    BuildExportTable();         /* fill in the .drectve section with -export's */
+    /* fill in the .drectve section with -export's */
+    BuildExportTable(&Exports);
 
     if (win32) {
-	/* add default value for @feat.00, this allows to 'link /safeseh' */
-	uint32_t n;
+        /* add default value for @feat.00, this allows to 'link /safeseh' */
+        uint32_t n;
 
-	saa_rewind(syms);
-	for (n = 0; n < nsyms; n++) {
+        saa_rewind(syms);
+        for (n = 0; n < nsyms; n++) {
             struct Symbol *sym = saa_rstruct(syms);
-	    if (sym->strpos == -1 && !strcmp("@feat.00",sym->name))
-		break;
-	}
-	if (n == nsyms)
-	    coff_deflabel("@feat.00",NO_SEG,1,0,NULL);
+            if (sym->strpos == -1 && !strcmp("@feat.00",sym->name))
+                break;
+        }
+        if (n == nsyms)
+            coff_deflabel("@feat.00", NO_SEG, 1, 0, NULL);
     }
 
     /*
@@ -840,16 +881,17 @@ static void coff_write(void)
      * Output the COFF header.
      */
     if (win64)
-        fwriteint16_t(0x8664, coffp);  /* MACHINE_x86-64 */
+        i = IMAGE_FILE_MACHINE_AMD64;
     else
-        fwriteint16_t(0x014C, coffp);  /* MACHINE_i386 */
-    fwriteint16_t(nsects, coffp); /* number of sections */
-    fwriteint32_t(time(NULL), coffp);      /* time stamp */
-    fwriteint32_t(sympos, coffp);
-    fwriteint32_t(nsyms + initsym, coffp);
-    fwriteint16_t(0, coffp);      /* no optional header */
+        i = IMAGE_FILE_MACHINE_I386;
+    fwriteint16_t(i,                    ofile); /* machine type */
+    fwriteint16_t(nsects,               ofile); /* number of sections */
+    fwriteint32_t(time(NULL),           ofile); /* time stamp */
+    fwriteint32_t(sympos,               ofile);
+    fwriteint32_t(nsyms + initsym,      ofile);
+    fwriteint16_t(0,                    ofile); /* no optional header */
     /* Flags: 32-bit, no line numbers. Win32 doesn't even bother with them. */
-    fwriteint16_t((win32 | win64) ? 0 : 0x104, coffp);
+    fwriteint16_t((win32 | win64) ? 0 : 0x104, ofile);
 
     /*
      * Output the section headers.
@@ -867,7 +909,7 @@ static void coff_write(void)
      */
     for (i = 0; i < nsects; i++)
         if (sects[i]->data) {
-            saa_fpwrite(sects[i]->data, coffp);
+            saa_fpwrite(sects[i]->data, ofile);
             coff_write_relocs(sects[i]);
         }
 
@@ -875,8 +917,8 @@ static void coff_write(void)
      * Output the symbol and string tables.
      */
     coff_write_symbols();
-    fwriteint32_t(strslen + 4, coffp);     /* length includes length count */
-    saa_fpwrite(strs, coffp);
+    fwriteint32_t(strslen + 4, ofile);     /* length includes length count */
+    saa_fpwrite(strs, ofile);
 }
 
 static void coff_section_header(char *name, int32_t vsize,
@@ -889,16 +931,17 @@ static void coff_section_header(char *name, int32_t vsize,
 
     memset(padname, 0, 8);
     strncpy(padname, name, 8);
-    fwrite(padname, 8, 1, coffp);
-    fwriteint32_t(0, coffp);	/* Virtual size field - set to 0 or vsize */
-    fwriteint32_t(0L, coffp);      /* RVA/offset - we ignore */
-    fwriteint32_t(datalen, coffp);
-    fwriteint32_t(datapos, coffp);
-    fwriteint32_t(relpos, coffp);
-    fwriteint32_t(0L, coffp);      /* no line numbers - we don't do 'em */
-    fwriteint16_t(nrelocs, coffp);
-    fwriteint16_t(0, coffp);      /* again, no line numbers */
-    fwriteint32_t(flags, coffp);
+    fwrite(padname, 8, 1, ofile);
+
+    fwriteint32_t(0,            ofile); /* Virtual size field - set to 0 or vsize */
+    fwriteint32_t(0L,           ofile); /* RVA/offset - we ignore */
+    fwriteint32_t(datalen,      ofile);
+    fwriteint32_t(datapos,      ofile);
+    fwriteint32_t(relpos,       ofile);
+    fwriteint32_t(0L,           ofile); /* no line numbers - we don't do 'em */
+    fwriteint16_t(nrelocs,      ofile);
+    fwriteint16_t(0,            ofile); /* again, no line numbers */
+    fwriteint32_t(flags,        ofile);
 }
 
 static void coff_write_relocs(struct Section *s)
@@ -906,12 +949,12 @@ static void coff_write_relocs(struct Section *s)
     struct Reloc *r;
 
     for (r = s->head; r; r = r->next) {
-        fwriteint32_t(r->address, coffp);
+        fwriteint32_t(r->address, ofile);
         fwriteint32_t(r->symbol + (r->symbase == REAL_SYMBOLS ? initsym :
-                                r->symbase == ABS_SYMBOL ? initsym - 1 :
-                                r->symbase == SECT_SYMBOLS ? 2 : 0),
-                   coffp);
-	fwriteint16_t(r->type, coffp);
+                                   r->symbase == ABS_SYMBOL   ? initsym - 1 :
+                                   r->symbase == SECT_SYMBOLS ? 2 : 0),
+                      ofile);
+        fwriteint16_t(r->type, ofile);
     }
 }
 
@@ -923,16 +966,18 @@ static void coff_symbol(char *name, int32_t strpos, int32_t value,
     if (name) {
         memset(padname, 0, 8);
         strncpy(padname, name, 8);
-        fwrite(padname, 8, 1, coffp);
+        fwrite(padname, 8, 1, ofile);
     } else {
-        fwriteint32_t(0, coffp);
-        fwriteint32_t(strpos, coffp);
+        fwriteint32_t(0, ofile);
+        fwriteint32_t(strpos, ofile);
     }
-    fwriteint32_t(value, coffp);
-    fwriteint16_t(section, coffp);
-    fwriteint16_t(type, coffp);
-    fputc(storageclass, coffp);
-    fputc(aux, coffp);
+
+    fwriteint32_t(value,        ofile);
+    fwriteint16_t(section,      ofile);
+    fwriteint16_t(type,         ofile);
+
+    fputc(storageclass, ofile);
+    fputc(aux, ofile);
 }
 
 static void coff_write_symbols(void)
@@ -946,7 +991,7 @@ static void coff_write_symbols(void)
     coff_symbol(".file", 0L, 0L, -2, 0, 0x67, 1);
     memset(filename, 0, 18);
     strncpy(filename, coff_infile, 18);
-    fwrite(filename, 18, 1, coffp);
+    fwrite(filename, 18, 1, ofile);
 
     /*
      * The section records, with their auxiliaries.
@@ -955,9 +1000,9 @@ static void coff_write_symbols(void)
 
     for (i = 0; i < (uint32_t) nsects; i++) {
         coff_symbol(sects[i]->name, 0L, 0L, i + 1, 0, 3, 1);
-        fwriteint32_t(sects[i]->len, coffp);
-        fwriteint16_t(sects[i]->nrelocs, coffp);
-        fwrite(filename, 12, 1, coffp);
+        fwriteint32_t(sects[i]->len,    ofile);
+        fwriteint16_t(sects[i]->nrelocs,ofile);
+        fwrite(filename, 12, 1, ofile);
     }
 
     /*
@@ -977,21 +1022,47 @@ static void coff_write_symbols(void)
     }
 }
 
+static void coff_sectalign(int32_t seg, unsigned int value)
+{
+    struct Section *s = NULL;
+    uint32_t align;
+    int i;
+
+    for (i = 0; i < nsects; i++) {
+        if (sects[i]->index == seg) {
+            s = sects[i];
+            break;
+        }
+    }
+
+    if (!s || !is_power2(value))
+        return;
+
+    /* DOS has limitation on 64 bytes */
+    if (!(win32 | win64) && value > 64)
+        return;
+
+    align = (s->flags & IMAGE_SCN_ALIGN_MASK);
+    value = coff_sectalign_flags(value);
+    if (value > align)
+        s->flags = (s->flags & ~IMAGE_SCN_ALIGN_MASK) | value;
+}
+
 static int32_t coff_segbase(int32_t segment)
 {
     return segment;
 }
 
-static void coff_std_filename(char *inname, char *outname, efunc error)
+static void coff_std_filename(char *inname, char *outname)
 {
     strcpy(coff_infile, inname);
-    standard_extension(inname, outname, ".o", error);
+    standard_extension(inname, outname, ".o");
 }
 
-static void coff_win32_filename(char *inname, char *outname, efunc error)
+static void coff_win32_filename(char *inname, char *outname)
 {
     strcpy(coff_infile, inname);
-    standard_extension(inname, outname, ".obj", error);
+    standard_extension(inname, outname, ".obj");
 }
 
 extern macros_t coff_stdmac[];
@@ -1002,7 +1073,7 @@ static int coff_set_info(enum geninfo type, char **val)
     (void)val;
     return 0;
 }
-#endif                          /* defined(OF_COFF) || defined(OF_WIN32) */
+#endif /* defined(OF_COFF) || defined(OF_WIN32) */
 
 #ifdef OF_COFF
 
@@ -1018,6 +1089,7 @@ struct ofmt of_coff = {
     coff_out,
     coff_deflabel,
     coff_section_names,
+    coff_sectalign,
     coff_segbase,
     coff_directives,
     coff_std_filename,
@@ -1040,6 +1112,7 @@ struct ofmt of_win32 = {
     coff_out,
     coff_deflabel,
     coff_section_names,
+    coff_sectalign,
     coff_segbase,
     coff_directives,
     coff_win32_filename,
@@ -1062,6 +1135,7 @@ struct ofmt of_win64 = {
     coff_out,
     coff_deflabel,
     coff_section_names,
+    coff_sectalign,
     coff_segbase,
     coff_directives,
     coff_win32_filename,
